@@ -10,12 +10,10 @@ from typing import List
 from typing import Union
 from typing import Optional
 from typing import NamedTuple
-from typing import Dict
-from typing import Protocol
-from typing import Any
+from typing import TYPE_CHECKING
 
 from liquid.exceptions import LiquidSyntaxError, Error
-from liquid.mode import Mode, error
+from liquid.mode import error
 
 from liquid import expression
 from liquid import ast
@@ -63,25 +61,16 @@ from liquid.token import TOKEN_GE
 from liquid.token import reverse_operators
 from liquid.token import Token
 
-
-class Tag(Protocol):
-
-    block: bool
-
-    def get_node(self: Any, stream: TokenStream) -> ast.Node:
-        ...
-
-
-class Env(Protocol):
-
-    tags: Dict[str, Tag]
-    mode: Mode
+if TYPE_CHECKING:
+    from liquid import Environment
 
 
 class Parser:
+    """A Liquid template parser. Create a parser tree from a stream of tokens."""
+
     __slots__ = ("tags", "mode", "illegal", "literal", "statement")
 
-    def __init__(self, env: Env):
+    def __init__(self, env: Environment):
         self.tags = env.tags
         self.mode = env.mode
 
@@ -90,11 +79,13 @@ class Parser:
         self.statement = self.tags[TOKEN_STATEMENT]
 
     def parse(self, stream: TokenStream) -> ast.ParseTree:
+        """Parse the given stream of tokens into a tree."""
         root = ast.ParseTree()
 
         while stream.current.type != TOKEN_EOF:
             try:
-                if node := self.parse_statement(stream):
+                node = self.parse_statement(stream)
+                if node:
                     root.statements.append(node)
             except Error as err:
                 error(self.mode, err, linenum=stream.current.linenum)
@@ -104,7 +95,6 @@ class Parser:
         return root
 
     def parse_statement(self, stream: TokenStream) -> ast.Node:
-        """"""
         if stream.current.test(TOKEN_STATEMENT):
             node = self.statement.get_node(stream)
         elif stream.current.test(TOKEN_TAG):
@@ -219,71 +209,6 @@ IDENTIFIER_TOKENS = frozenset(
 )
 
 
-def peek_precedence(stream: TokenStream) -> Precedence:
-    """Return the precedence of the next token in the stream."""
-    try:
-        precedence = PRECEDENCES[stream.peek.type]
-    except KeyError as err:
-        raise LiquidSyntaxError(f"unknown operator '{stream.peek.value}'") from err
-    return precedence
-
-
-def current_precedence(stream: TokenStream) -> Precedence:
-    """Return the precedence of the current token in the stream."""
-    precedence = PRECEDENCES.get(stream.current.type, Precedence.LOWEST)
-    if stream.current.type in RIGHT_ASSOCIATIVE:
-        precedence -= 1
-    return Precedence(precedence)
-
-
-def parse_filters(stream: TokenStream) -> List[expression.Filter]:
-    """Keep reading filters from the token stream until end of expression."""
-    filters = []
-    while stream.current.type != TOKEN_EOF:
-        filters.append(parse_filter(stream))
-    return filters
-
-
-def parse_filter(stream: TokenStream) -> expression.Filter:
-    """Read a single filter with optional arguments from the token stream."""
-    # Eat the pipe token
-    expect(stream, TOKEN_PIPE)
-    stream.next_token()
-
-    # Read the filter's name
-    expect(stream, TOKEN_IDENTIFIER)
-    filter_name = stream.current.value
-    stream.next_token()
-
-    #
-    args = []
-    kwargs = {}
-
-    if stream.current.type == TOKEN_COLON:
-        stream.next_token()
-
-        while stream.current.type not in (TOKEN_PIPE, TOKEN_EOF):
-            if stream.peek.type == TOKEN_COLON:
-                # A named parameter
-                param_name = stream.current.value
-                stream.next_token()
-
-                # Eat colon
-                stream.next_token()
-
-                param_value = parse_expression(stream)
-                kwargs[param_name] = param_value
-            else:
-                # A positional argument/parameter
-                args.append(parse_expression(stream))
-            stream.next_token()
-
-            if stream.current.type == TOKEN_COMMA:
-                stream.next_token()
-
-    return expression.Filter(filter_name, args, kwargs)
-
-
 def parse_boolean(stream: TokenStream) -> expression.Boolean:
     """Read a boolean literal (true or false) from the token stream."""
     if stream.current.type == TOKEN_TRUE:
@@ -438,131 +363,6 @@ def parse_range_option(stream: TokenStream) -> RangeOption:
     return opt
 
 
-def parse_prefix_expression(stream: TokenStream) -> expression.PrefixExpression:
-    tok = stream.current
-    stream.next_token()
-
-    exp = expression.PrefixExpression(
-        tok.value,
-        right=parse_expression(stream, precedence=Precedence.PREFIX),
-    )
-
-    return exp
-
-
-def parse_infix_expression(
-    stream: TokenStream, left: expression.Expression
-) -> expression.InfixExpression:
-    tok = stream.current
-    precedence = current_precedence(stream)
-    stream.next_token()
-
-    exp = expression.InfixExpression(
-        left=left,
-        operator=tok.value,
-        right=parse_expression(stream, precedence),
-    )
-    return exp
-
-
-prefix_funcs = {
-    TOKEN_FALSE: parse_boolean,
-    TOKEN_TRUE: parse_boolean,
-    TOKEN_NIL: parse_nil,
-    TOKEN_EMPTY: parse_empty,
-    TOKEN_NEGATIVE: parse_prefix_expression,
-    TOKEN_STRING: parse_string_literal,
-    TOKEN_INTEGER: parse_integer_literal,
-    TOKEN_FLOAT: parse_float_literal,
-    TOKEN_IDENTIFIER: parse_identifier,
-}
-
-infix_funcs = {
-    TOKEN_EQ: parse_infix_expression,
-    TOKEN_OR: parse_infix_expression,
-    TOKEN_AND: parse_infix_expression,
-    TOKEN_LT: parse_infix_expression,
-    TOKEN_GT: parse_infix_expression,
-    TOKEN_NE: parse_infix_expression,
-    TOKEN_LG: parse_infix_expression,
-    TOKEN_LE: parse_infix_expression,
-    TOKEN_GE: parse_infix_expression,
-    TOKEN_CONTAINS: parse_infix_expression,
-}
-
-
-def parse_expression(
-    stream: TokenStream,
-    precedence: Precedence = Precedence.LOWEST,
-) -> expression.Expression:
-    """Parse a literal or logical expression."""
-    prefix = prefix_funcs.get(stream.current.type)
-    if prefix is None:
-        operator = reverse_operators.get(stream.current.type, stream.current.value)
-        raise LiquidSyntaxError(f"unknown prefix operator '{operator}'")
-
-    left = prefix(stream)
-
-    while (
-        stream.peek.type != TOKEN_EOF
-        and stream.peek.type not in (TOKEN_PIPE, TOKEN_COMMA)
-        and precedence < peek_precedence(stream)
-    ):
-        infix = infix_funcs.get(stream.peek.type)
-        if infix is None:
-            return left
-
-        stream.next_token()
-        left = infix(stream, left)
-
-    return left
-
-
-def parse_filtered_expression(stream: TokenStream) -> expression.FilteredExpression:
-    """Parse a statement (output) expression with optional filters.
-
-    A statement expression is assumed to start with exactly one identifier, followed by
-    zero or more filters.
-    """
-    expr = parse_expression(stream)
-    stream.next_token()
-    filters = parse_filters(stream)
-    return expression.FilteredExpression(expr, filters)
-
-
-def parse_boolean_expression(stream: TokenStream) -> expression.Expression:
-    """Parse a liquid expression that evaluates to a Boolean value.
-
-    This is primarily used by control flow tags like `if`, `unless` and `case`/`when`.
-
-    Logical operators (`and` and `or`) are right associative. They are evaluated from
-    right to left. The use of parentheses to control operator precedence is not allowed.
-    """
-    if stream.current.type == TOKEN_EOF:
-        # Empty expression.
-        return expression.Nil()
-    return expression.BooleanExpression(expression=parse_expression(stream))
-
-
-def parse_assignment_expression(stream: TokenStream) -> expression.AssignmentExpression:
-    """Parse a liquid assignment expression, as one might find in an `assign` tag.
-
-    This is essentially the same as a parse_filtered_expression, but with
-    an additional name to bind the expression result to.
-    """
-    expect(stream, TOKEN_IDENTIFIER)
-    expect_peek(stream, TOKEN_ASSIGN)
-
-    name = parse_unchained_identifier(stream)
-
-    stream.next_token()
-    stream.next_token()  # Eat ASSIGN TOKEN
-
-    expr = parse_filtered_expression(stream)
-
-    return expression.AssignmentExpression(name=str(name), expression=expr)
-
-
 def parse_loop_expression(stream: TokenStream) -> expression.LoopExpression:
     """Parse a liquid loop expression, as one might find in a `for` tag.
 
@@ -625,6 +425,228 @@ def parse_loop_expression(stream: TokenStream) -> expression.LoopExpression:
     return expr
 
 
-@lru_cache
-def get_parser(env: Env) -> Parser:
+class ExpressionParser:
+    """The standard expression parser.
+
+    Custom, non "standard" tags can subclass `ExpressionParser` to change expression
+    parsing behavior. Such as adding a logical `NOT` operator.
+    """
+
+    def __init__(self):
+        self.precedences = PRECEDENCES.copy()
+        self.right_associative = RIGHT_ASSOCIATIVE.copy()
+
+        self.prefix_funcs = {
+            TOKEN_FALSE: parse_boolean,
+            TOKEN_TRUE: parse_boolean,
+            TOKEN_NIL: parse_nil,
+            TOKEN_EMPTY: parse_empty,
+            TOKEN_NEGATIVE: self.parse_prefix_expression,
+            TOKEN_STRING: parse_string_literal,
+            TOKEN_INTEGER: parse_integer_literal,
+            TOKEN_FLOAT: parse_float_literal,
+            TOKEN_IDENTIFIER: parse_identifier,
+        }
+
+        self.infix_funcs = {
+            TOKEN_EQ: self.parse_infix_expression,
+            TOKEN_OR: self.parse_infix_expression,
+            TOKEN_AND: self.parse_infix_expression,
+            TOKEN_LT: self.parse_infix_expression,
+            TOKEN_GT: self.parse_infix_expression,
+            TOKEN_NE: self.parse_infix_expression,
+            TOKEN_LG: self.parse_infix_expression,
+            TOKEN_LE: self.parse_infix_expression,
+            TOKEN_GE: self.parse_infix_expression,
+            TOKEN_CONTAINS: self.parse_infix_expression,
+        }
+
+    def peek_precedence(self, stream: TokenStream) -> Precedence:
+        """Return the precedence of the next token in the stream."""
+        try:
+            precedence = self.precedences[stream.peek.type]
+        except KeyError as err:
+            raise LiquidSyntaxError(f"unknown operator '{stream.peek.value}'") from err
+        return precedence
+
+    def current_precedence(self, stream: TokenStream) -> Precedence:
+        """Return the precedence of the current token in the stream."""
+        precedence = self.precedences.get(stream.current.type, Precedence.LOWEST)
+        if stream.current.type in self.right_associative:
+            precedence -= 1
+        return Precedence(precedence)
+
+    def parse_filters(self, stream: TokenStream) -> List[expression.Filter]:
+        """Keep reading filters from the token stream until end of expression."""
+        filters = []
+        while stream.current.type != TOKEN_EOF:
+            filters.append(self.parse_filter(stream))
+        return filters
+
+    def parse_filter(self, stream: TokenStream) -> expression.Filter:
+        """Read a single filter with optional arguments from the token stream."""
+        # Eat the pipe token
+        expect(stream, TOKEN_PIPE)
+        stream.next_token()
+
+        # Read the filter's name
+        expect(stream, TOKEN_IDENTIFIER)
+        filter_name = stream.current.value
+        stream.next_token()
+
+        #
+        args = []
+        kwargs = {}
+
+        if stream.current.type == TOKEN_COLON:
+            stream.next_token()
+
+            while stream.current.type not in (TOKEN_PIPE, TOKEN_EOF):
+                if stream.peek.type == TOKEN_COLON:
+                    # A named parameter
+                    param_name = stream.current.value
+                    stream.next_token()
+
+                    # Eat colon
+                    stream.next_token()
+
+                    param_value = self.parse_expression(stream)
+                    kwargs[param_name] = param_value
+                else:
+                    # A positional argument/parameter
+                    args.append(self.parse_expression(stream))
+                stream.next_token()
+
+                if stream.current.type == TOKEN_COMMA:
+                    stream.next_token()
+
+        return expression.Filter(filter_name, args, kwargs)
+
+    def parse_prefix_expression(
+        self, stream: TokenStream
+    ) -> expression.PrefixExpression:
+        tok = stream.current
+        stream.next_token()
+
+        exp = expression.PrefixExpression(
+            tok.value,
+            right=self.parse_expression(stream, precedence=Precedence.PREFIX),
+        )
+
+        return exp
+
+    def parse_infix_expression(
+        self,
+        stream: TokenStream,
+        left: expression.Expression,
+    ) -> expression.InfixExpression:
+        tok = stream.current
+        precedence = self.current_precedence(stream)
+        stream.next_token()
+
+        exp = expression.InfixExpression(
+            left=left,
+            operator=tok.value,
+            right=self.parse_expression(stream, precedence),
+        )
+        return exp
+
+    def parse_expression(
+        self,
+        stream: TokenStream,
+        precedence: Precedence = Precedence.LOWEST,
+    ) -> expression.Expression:
+        """Parse a literal or logical expression."""
+        prefix = self.prefix_funcs.get(stream.current.type)
+        if prefix is None:
+            operator = reverse_operators.get(stream.current.type, stream.current.value)
+            raise LiquidSyntaxError(f"unknown prefix operator '{operator}'")
+
+        left = prefix(stream)
+
+        while (
+            stream.peek.type != TOKEN_EOF
+            and stream.peek.type not in (TOKEN_PIPE, TOKEN_COMMA)
+            and precedence < self.peek_precedence(stream)
+        ):
+            infix = self.infix_funcs.get(stream.peek.type)
+            if infix is None:
+                return left
+
+            stream.next_token()
+            left = infix(stream, left)
+
+        return left
+
+    def parse_filtered_expression(
+        self, stream: TokenStream
+    ) -> expression.FilteredExpression:
+        """Parse a statement (output) expression with optional filters.
+
+        A statement expression is assumed to start with exactly one identifier, followed
+        by zero or more filters.
+        """
+        expr = self.parse_expression(stream)
+        stream.next_token()
+        filters = self.parse_filters(stream)
+        return expression.FilteredExpression(expr, filters)
+
+    def parse_boolean_expression(self, stream: TokenStream) -> expression.Expression:
+        """Parse a liquid expression that evaluates to a Boolean value.
+
+        This is primarily used by control flow tags like `if`, `unless` and `case`/
+        `when`.
+
+        Logical operators (`and` and `or`) are right associative. They are evaluated
+        from right to left. The use of parentheses to control operator precedence is
+        not allowed.
+        """
+        if stream.current.type == TOKEN_EOF:
+            # Empty expression.
+            return expression.Nil()
+        return expression.BooleanExpression(expression=self.parse_expression(stream))
+
+    def parse_assignment_expression(
+        self, stream: TokenStream
+    ) -> expression.AssignmentExpression:
+        """Parse a liquid assignment expression, as one might find in an `assign` tag.
+
+        This is essentially the same as a parse_filtered_expression, but with
+        an additional name to bind the expression result to.
+        """
+        expect(stream, TOKEN_IDENTIFIER)
+        expect_peek(stream, TOKEN_ASSIGN)
+
+        name = parse_unchained_identifier(stream)
+
+        stream.next_token()
+        stream.next_token()  # Eat ASSIGN TOKEN
+
+        expr = self.parse_filtered_expression(stream)
+
+        return expression.AssignmentExpression(name=str(name), expression=expr)
+
+
+# For backwards compatibility.
+STANDARD_EXPRESSION_PARSER = ExpressionParser()
+peek_precedence = STANDARD_EXPRESSION_PARSER.peek_precedence
+current_precedence = STANDARD_EXPRESSION_PARSER.current_precedence
+parse_filters = STANDARD_EXPRESSION_PARSER.parse_filters
+parse_filter = STANDARD_EXPRESSION_PARSER.parse_filter
+parse_prefix_expression = STANDARD_EXPRESSION_PARSER.parse_prefix_expression
+parse_infix_expression = STANDARD_EXPRESSION_PARSER.parse_infix_expression
+parse_expression = STANDARD_EXPRESSION_PARSER.parse_expression
+parse_filtered_expression = STANDARD_EXPRESSION_PARSER.parse_filtered_expression
+parse_boolean_expression = STANDARD_EXPRESSION_PARSER.parse_boolean_expression
+parse_assignment_expression = STANDARD_EXPRESSION_PARSER.parse_assignment_expression
+
+
+@lru_cache(maxsize=128)
+def get_parser(env: Environment) -> Parser:
     return Parser(env)
+
+
+@lru_cache(maxsize=128)
+def get_expression_parser(_: Environment) -> ExpressionParser:
+    # Future proofing.
+    return STANDARD_EXPRESSION_PARSER
